#version 430 core
in VertexData {
    vec3 position_world;
    vec3 normal_world;
    vec2 uv;
} vert;

// out vec4 color;
layout (location = 0)  out vec4 color;
layout (location = 1)  out vec4 bright;

uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular 
uniform float specularAlpha;
uniform sampler2D diffuseTexture;
uniform sampler2D bloomTexture;
uniform bool shadowDebug;
uniform bool postProcess;

uniform struct DirectionalLight {
    vec3 color;
    vec3 direction;
} dirL;

struct PointLight {
    vec3 color;
    vec3 position;
    vec3 attenuation;
};


uniform float far_plane;

#define NR_POINT_LIGHTS 3
uniform samplerCube shadowMap[NR_POINT_LIGHTS];
uniform PointLight pointL[NR_POINT_LIGHTS];

vec3 sampleOffsetDirections[20] = vec3[]
(
   vec3( 1,  1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1,  1,  1), 
   vec3( 1,  1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1,  1, -1),
   vec3( 1,  1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1,  1,  0),
   vec3( 1,  0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1,  0, -1),
   vec3( 0,  1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0,  1, -1)
);   

vec3 phong(PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir, float shadow)
{
    vec3 lightDir = normalize(light.position - fragPos);
    vec3 texColor = texture(diffuseTexture, vert.uv).rgb * light.color;

    // diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    // specular shading
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), 0.1); // shininess needs to be set by us in the future 

    // attenuation
    float distance    = length(light.position - fragPos);
    float attenuation = 1.0 / (light.attenuation.x + light.attenuation.y * distance + 
  			     light.attenuation.z * (distance * distance));    
    // combine results
    vec3 ambient  = attenuation * texColor;
    vec3 diffuse  = attenuation * texColor * diff;
    vec3 specular = attenuation * texColor * spec * specularAlpha;
    // TODO specular destroys our light range quite a bit 
    // return ambient * diffuse* specular * (1-shadow);
    return ambient * diffuse * (1-shadow);
}  

float ShadowCalculation(vec3 fragPos, PointLight light, samplerCube shadowMap)
{
    // get vector between fragment position and light position
    vec3 fragToLight = fragPos - light.position;

    // use the light to fragment vector to sample from the depth map  
    float currentDepth = length(fragToLight); 

    float shadow = 0.0;
    float bias   = 0.15;
    int samples  = 20;
    float viewDistance = length(camera_world - fragPos);
    //float diskRadius = 0.05;
    float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;  
    for(int i = 0; i < samples; ++i)
    {
        float closestDepth = texture(shadowMap, fragToLight + sampleOffsetDirections[i] * diskRadius).r;
        closestDepth *= far_plane;   // undo mapping [0;1]
        if (shadowDebug) {
        color.rgba = vec4(vec3(closestDepth / far_plane), 1.0);  
        }
        if(currentDepth - bias > closestDepth){
            shadow += 1.0;
        } 
    }
    shadow /= float(samples);  

    return shadow; 

}

void DepthCalculation(vec3 fragPos, PointLight light, samplerCube shadowMap)
{
    // get vector between fragment position and light position
    vec3 fragToLight = fragPos - light.position;

    // use the light to fragment vector to sample from the depth map  
    float currentDepth = length(fragToLight); 

    float shadow = 0.0;
    float bias   = 0.15;
    int samples  = 20;
    float viewDistance = length(camera_world - fragPos);
    //float diskRadius = 0.05;
    float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;  
    for(int i = 0; i < samples; ++i)
    {
        float closestDepth = texture(shadowMap, fragToLight + sampleOffsetDirections[i] * diskRadius).r;
        closestDepth *= far_plane;   // undo mapping [0;1]
        if (shadowDebug) {
           color.rgba = vec4(vec3(closestDepth / far_plane), 1.0);      
        }
        if(currentDepth - bias > closestDepth){
            shadow += 1.0;
        } 
    }
    shadow /= float(samples);  


}

void main() {
    vec3 n = normalize(vert.normal_world);
    vec3 v = normalize(camera_world - vert.position_world);

    vec3 texColor = texture(diffuseTexture, vert.uv).rgb;

    if (!shadowDebug) {
        color = vec4(texColor * materialCoefficients.x, 1); // ambient
    }

    // Potential TODO - Look into potential Light Setup 
    float shadow = 0.0;

    for (int i = 0; i < NR_POINT_LIGHTS; i++)
    {
        if (!shadowDebug)
        {
            shadow = ShadowCalculation(vert.position_world, pointL[i], shadowMap[i]);
            // Accumulate the phong shading using the accumulated shadow value
             color.rgb += phong(pointL[i], n, vert.position_world, pointL[i].position - vert.position_world, shadow);
       }
        else
        {
            DepthCalculation(vert.position_world, pointL[i], shadowMap[i]);
        }
    }
    float brightness =  dot(color.rgb, vec3(0.2126, 0.7152, 0.0722));
    if (brightness > 1.0 && postProcess) {
        bright = vec4(color.rgb, 1.0);
    }else {
        bright = vec4(0.0,0.0,0.0,1.0);

    }
}
